import html
import re
from naturalsize import replStrPassage

from .defaults import HELLO_WORLD as DEFAULT
from .commons import *
from .exceptions import InvalidIncludePhraseFiletypeError, StaticResourceUsageOutsideHeadError
from .beta import BETA
from .patterns import *

class PyHTML:
    def __init__(self, html: str = DEFAULT):
        self.html = html
        self.context = None

    def _replace_eval_phrase(self):
        """
        Scans the HTML for the evalPyHTML phrase and replaces it with the appropriate start and end phrases.
        """
        if self.html.startswith(EVAL_PYHTML):
            self.html = START_REPLACE + self.html[len(EVAL_PYHTML):]
        if self.html.endswith(EVAL_PYHTML):
            self.html = self.html[:-len(EVAL_PYHTML)] + END_REPLACE

    def _replace_modern_styling(self):
        """
        Scans the HTML for the modern styling phrase and replaces it with the appropriate CSS.
        """
        idx = self.html.find(MODERN_STYLING_PHRASE)
        if idx != -1:
            head_idx = self.html.find("</head>")
            if idx > head_idx:
                raise StaticResourceUsageOutsideHeadError()
            self.html = replStrPassage(idx, idx+len(MODERN_STYLING_PHRASE), self.html, MODERN_STYLING)

    def _replace_includes(self):
        """
        Scans the HTML for include phrases and replaces them with the appropriate resources.
        """
        for match in re.finditer(INCLUDE_PATTERN, self.html):
            idx, idxEnd = match.span()
            head_idx = self.html.find("</head>")
            if idx > head_idx:
                raise StaticResourceUsageOutsideHeadError()
            resources = match.group(1).split(",")
            setIn = ""
            for i in resources:
                i = i.strip()
                if i.endswith(".css"):
                    setIn += f"\t\t<link rel='stylesheet' href='{i}'/>\n"
                elif i.endswith(".js"):
                    setIn += f"\t\t<script src='{i}'></script>\n"
                elif i.endswith(".json"):
                    setIn += f"\t\t<link rel='manifest' href='{i}'/>\n"
                elif i.endswith("favicon.ico"):
                    setIn += f"\t\t<link rel='icon' href='{i}'/>\n"
                else:
                    raise InvalidIncludePhraseFiletypeError()
            self.html = self.html[:idx] + setIn + self.html[idxEnd:]

    def _replace_script_blocks(self):
        """
        Replaces script replacement phrases with JS scripts.
        """
        for match in re.finditer(SCRIPT_PATTERN, self.html, re.DOTALL):
            idx, idxEnd = match.span()
            script_content = match.group(1).strip()
            replacement = f"<script>{script_content}</script>"
            self.html = self.html[:idx] + replacement + self.html[idxEnd:]

    def _replace_style_blocks(self):
        """
        Replaces style replacement phrases with CSS styles.
        """
        for match in re.finditer(STYLE_PATTERN, self.html, re.DOTALL):
            idx, idxEnd = match.span()
            style_content = match.group(1).strip()
            replacement = f"<style>{style_content}</style>"
            self.html = self.html[:idx] + replacement + self.html[idxEnd:]

    def _replace_eval_blocks(self):
        """
        Replaces eval replacement phrases with evaluated Python expressions.
        WARNING: This method uses eval, which can execute arbitrary code.
        """
        if self.context is None:
            self.context = {}
        for match in re.finditer(EVAL_BLOCK_PATTERN, self.html, re.DOTALL):
            idx, idxEnd = match.span()
            code = match.group(1).strip()
            try:
                # eval für Ausdrücke, exec für Statements
                result = str(eval(code, {}, self.context))
            except Exception as e:
                result = f"<b>EvalError: {e}</b>"
            self.html = self.html[:idx] + result + self.html[idxEnd:]

    def _replace_if_blocks(self, context=None):
        """
        Replaces with html code based on the evaluation of conditions in if blocks.
        """
        if context is None:
            context = {}
        # re.DOTALL für mehrzeilige Blöcke
        while True:
            match = re.search(IF_BLOCK_PATTERN, self.html, re.DOTALL)
            if not match:
                break
            condition = match.group(1).strip()
            if_content = match.group(2)
            else_content = match.group(5) if match.group(5) is not None else ""
            try:
                if eval(condition, {}, context):
                    replacement = if_content
                else:
                    replacement = else_content
            except Exception as e:
                replacement = f"<b>IfEvalError: {e}</b>"
            self.html = self.html[:match.start()] + replacement + self.html[match.end():]

    def decode(self):
        """
        Decodes the HTML content by replacing specific phrases and applying modern styling.
        """
        self.html = self.html.strip()
        self._replace_eval_phrase()
        self._replace_modern_styling()
        self._replace_script_blocks()
        self._replace_style_blocks()
        if BETA.value:
            self._replace_includes()
            self._replace_eval_blocks()
            self._replace_if_blocks()

    def decoded(self, cacheDecoded: bool = False) -> str:
        """
        Returns the decoded HTML content.
        """
        backup = self.html
        self.decode()
        _return = self.html
        if not cacheDecoded:
            self.html = backup
        return _return

    def decodedContext(self, context: dict, cacheDecoded: bool = False) -> str:
        """
        Returns the decoded HTML content with the provided context.
        """
        self.context = context
        return self.decoded(cacheDecoded)


START_REPLACE = """<!DOCTYPE html>
<html lang="de-de">
    <head>
        <meta charset="UTF-8"/>
        <meta name="viewport" content="width=device-width, initial-scale=1.0"/>
        <meta name="announcement" content="This is a PyWSGIRef generated page."/>
        """

END_REPLACE = """
        <p style="font-size: 7px; text-align: center;">Generated by PyWSGIRef</p>
    </body>
</html>"""

MODERN_STYLING = """<link rel='stylesheet' href='https://leanderkafemann.github.io/PyWSGIRef-Backend/styling/modernStyling.css'/>"""


"""
Main method creating simple server example...
"""

from PyWSGIRef import *

def main():
    """
    Main function to set up and run the PyWSGIRef server.
    """
    # add Schablone 'Hallo Welt' as main
    addSchablone("main", MAIN_HTML)

    # set up application object
    def contentGenerator(path: str) -> str:
        """
        Serves as the main WSGI application.
        """
        match path:
            case "/":
                content = SCHABLONEN["main"].decoded().format(about()["Version"])
            case "/hello":
                content = HELLO_WORLD
            case "/shutdown":
                content = SHUTDOWN_HTML
                server.shutdown()
            case _:
                content = ERROR
        return content

    # make the application object
    application = makeApplicationObject(contentGenerator)

    # set up server
    server = setUpServer(application)

    # Note: This code is intended to be run as a script, not as a module.
    print("Successfully started WSGI server on port 8000.")

    # start serving
    server.serve_forever()


HELLO_WORLD = """<!DOCTYPE html>
<html lang"de">
    <body>
        <h1>Hallo Welt!</h1>
    </body>
<html>"""

ERROR = """<!DOCTYPE html>
<html lang"de">
    <body>
        <h1>Fehler!</h1>
        <p>Es ist ein Fehler aufgetreten.</p>
    </body>
</html>"""

MAIN_HTML = """<!DOCTYPE html>
<html lang="de">
    <head>
        <meta charset="UTF-8" />
        <meta name="viewport" content="width=device-width, initial-scale=1.0" />
        <title>PyWSGIRef Server</title>
    </head>
    <body>
        <h1>Willkommen zum PyWSGIRef Server!</h1>
        <p>Dies ist die Hauptseite des Servers.</p>
        <p>Sie nutzen Version {} von PyWSGIRef.</p>
    </body>
</html>"""

SHUTDOWN_HTML = """<!DOCTYPE html>
<html lang="de">
    <head>
        <meta charset="UTF-8" />
        <meta name="viewport" content="width=device-width, initial-scale=1.0" />
        <title>Server Shutdown</title>
    <body>
        <h1>Server Shutdown</h1>
        <p>
            Der Server wird jetzt heruntergefahren.
        </p>
    </body>
</html>"""


EVAL_PYHTML = "<{{evalPyHTML}}>"

MODERN_STYLING_PHRASE = "<{{evalPyHTML-modernStyling: true}}>"

INCLUDE_PATTERN = r"<\{\{evalPyHTML-include: (.*?) :include-\}\}\>"

SCRIPT_PATTERN = r"<\{\{evalPyHTML-script: (.*?) :script-\}\}\>"

STYLE_PATTERN = r"<\{\{evalPyHTML-style: (.*?) :style-\}\}\>"

EVAL_BLOCK_PATTERN = r"<\{\{evalPyHTML-eval: (.*?) :eval-\}\}\>"

IF_BLOCK_PATTERN = r"<\{\{evalPyHTML-if: (.*?)\}\}\>(.*?)((<\{\{evalPyHTML-else\}\}\>(.*?))?)<\{\{evalPyHTML-endif\}\}\>"


"""
PYWSGIREF
"""
from typing import Callable
from wsgiref.simple_server import make_server, WSGIServer
from cgi import FieldStorage

from .exceptions import *
from .pyhtml import PyHTML
from .defaults import *
from .templateDict import TemplateDict
from .beta import BETA
from .loadContent import *
from .finished import finished

def about():
    """
    Returns information about your release and other projects by Leander Kafemann
    """
    return {"Version": (1, 1, 11), "Author": "Leander Kafemann", "date": "25.06.2025",\
            "recommend": ("Buero by LK",  "pyimager by LK"), "feedbackTo": "leander@kafemann.berlin"}

SCHABLONEN = TemplateDict()

def addSchablone(name: str, content: str):
    """
    Adds a template to the SCHABLONEN dictionary.
    """
    global SCHABLONEN
    if finished.value:
        raise ServerAlreadyGeneratedError()
    SCHABLONEN[name] = PyHTML(content)

def makeApplicationObject(contentGeneratingFunction: Callable, advanced: bool = False, setAdvancedHeaders: bool = False) -> Callable:
    """
    Returns a WSGI application object based on your contentGeneratingFunction.
    The contentGeneratingFunction should take a single argument (the path) and return the content as a string.
    If advanced is True, the contentGeneratingFunction will receive a FieldStorage object as the second argument.
    If setAdvancedHeaders is True, it will allow you to set advanced headers for the response.
    """
    if not callable(contentGeneratingFunction):
        raise InvalidCallableError()
    def simpleApplication(environ, start_response) -> list:
        """
        A simple WSGI application object that serves as a template.
        """
        type_ = "text/html" 
        status = "200 OK"
        if advanced:
            storage = FieldStorage(fp=environ.get("wsgi.input"), environ=environ, keep_blank_values=True)
            if setAdvancedHeaders:
                content, type_, status = contentGeneratingFunction(environ["PATH_INFO"], storage)
            else:
                content = contentGeneratingFunction(environ["PATH_INFO"], storage)
        else:
            if setAdvancedHeaders:
                raise AdvancedHeadersWithoutAdvancedModeError()
            content = contentGeneratingFunction(environ["PATH_INFO"])
        headers = [("Content-Type", type_),
                   ("Content-Length", str(len(content))),
                   ('Access-Control-Allow-Origin', '*')]
        start_response(status, headers)
        return [content.encode("utf-8")]
    return simpleApplication

def setUpServer(application: Callable, port: int = 8000) -> WSGIServer:
    """
    Creates a WSGI server.
    No additional Schablonen can be loaded from the web.
    """
    finished.set_true()
    server = make_server('', port, application)
    return server


class ServerAlreadyGeneratedError(Exception):
    """
    Raised when a server is already generated,
    so that further modifications are not allowed.
    """
    def __init__(self, message="Server has already been generated."):
        self.message = message
        super().__init__(self.message)

class InvalidFiletypeError(Exception):
    """
    Raised when an invalid file type (not PyHTML) is encountered.
    """
    def __init__(self, message="Invalid file type. Only .pyhtml files are allowed."):
        self.message = message
        super().__init__(self.message)

class InvalidCallableError(Exception):
    """
    Raised when a callable is expected but not provided.
    """
    def __init__(self, message="Expected a callable object."):
        self.message = message
        super().__init__(self.message)

class AdvancedHeadersWithoutAdvancedModeError(Exception):
    """
    Raised when advanced headers are set without advanced mode being enabled.
    """
    def __init__(self, message="Advanced headers cannot be set without enabling advanced mode."):
        self.message = message
        super().__init__(self.message)

class InvalidIncludePhraseFiletypeError(InvalidFiletypeError):
    """
    Raised when an invalid file type is encountered in the include phrase of a PyHTML file.
    """
    def __init__(self, message="Invalid file type in include phrase. Only .css, .js, and .json files are allowed."):
        self.message = message
        super().__init__(self.message)

class StaticResourceUsageOutsideHeadError(Exception):
    """
    Raised when a static resource is tried to include outside the head term of a PyHTML file.
    """
    def __init__(self, message="Static resource usage out of <head> term detected."):
        self.message = message
        super().__init__(self.message)

class BooleanAlreadyTrueError(Exception):
    """
    Raised when trying to set a boolean that is already True.
    """
    def __init__(self, message="Boolean value is already True."):
        self.message = message
        super().__init__(self.message)#

class BetaAlreadyEnabledError(BooleanAlreadyTrueError):
    """
    Raised when trying to enable beta mode that is already enabled.
    """
    def __init__(self, message="Beta mode is already enabled."):
        self.message = message
        super().__init__(self.message)


"""
BETA mode...
"""

from .exceptions import BetaAlreadyEnabledError

class beta:
    """
    BETA mode for PyWSGIRef.
    """
    def __init__(self):
        self._beta = False

    @property
    def value(self) -> bool:
        return self._beta

    def enable(self):
        """
        Enables BETA mode.
        """
        if self._beta:
            raise BetaAlreadyEnabledError()
        self._beta = True

BETA = beta()

def enableBetaMode():
    """
    Enables BETA mode.
    """
    BETA.enable()
    print("BETA mode enabled.")


from .exceptions import ServerAlreadyGeneratedError
from .pyhtml import PyHTML
from .finished import OneWayBoolean

class TemplateDict:
	def __init__(self):
		self.dictionary = {}
		self.locked = OneWayBoolean()

	def __getitem__(self, key: str) -> PyHTML:
		return self.dictionary[key]

	def __setitem__(self, key: str, value: PyHTML):
		if not isinstance(value, PyHTML):
			raise TypeError("Value must be an instance of PyHTML.")
		if not isinstance(key, str):
			raise TypeError("Key must be a string.")
		if self.locked.value:
			raise ServerAlreadyGeneratedError("Cannot modify TemplateDict after it has been locked.")
		self.dictionary[key] = value

	def __contains__(self, key: str):
		return key in self.dictionary

	def __repr__(self):
		return f"TemplateDict({self.dictionary})"

	def keys(self) -> list:
		return list(self.dictionary.keys())

	def values(self) -> list:
		return list(self.dictionary.values())

	def items(self) -> list:
		return list(self.dictionary.items())


"""
Content loading helper.
"""
import requests

from .exceptions import *
from .finished import finished

def loadFromWeb(url: str, data: dict = {}) -> str:
    """
    Loads content from the given URL with the given data.
    """
    if finished.value:
        raise ServerAlreadyGeneratedError()
    if not url.endswith(".pyhtml"):
        raise InvalidFiletypeError()
    rq = requests.post(url, data).content
    return rq.decode()

def loadFromFile(filename: str) -> str:
    """
    Loads a file from the given filename.
    """
    if finished.value:
        raise ServerAlreadyGeneratedError()
    if not filename.endswith(".pyhtml"):
        raise InvalidFiletypeError()
    with open(filename, "r", encoding="utf-8") as f:
        content = f.read()
    return content


"""
Creates finished object to indicate if the server has been generated.
"""

from .exceptions import BooleanAlreadyTrueError

class OneWayBoolean:
	def __init__(self):
		self._value = False

	@property
	def value(self) -> bool:
		return self._value

	def set_true(self):
		"""
		Setzt den Wert auf True, wenn er False ist.
		"""
		if not self._value:
			self._value = True
		else:
			raise BooleanAlreadyTrueError()

finished = OneWayBoolean()